import tempfile
import logging
from io import StringIO
from os.path import exists
from shutil import move as rename_pdb_file
from itertools import product, combinations

from openbabel import openbabel as ob
from openbabel.pybel import readfile
from openbabel.pybel import Molecule as PybelWrapper

from rdkit.Chem import MolFromMolBlock, MolFromMolFile, SanitizeFlags, SanitizeMol, MolToMolFile

from luna.util.file import is_directory_valid, new_unique_filename, remove_files
from luna.util.default_values import ENTRY_SEPARATOR, OPENBABEL
from luna.MyBio.selector import AtomSelector
from luna.MyBio.PDB.PDBList import PDBList
from luna.MyBio.PDB.PDBIO import Select
from luna.MyBio.PDB import Selection
from luna.MyBio.PDB.PDBIO import PDBIO
from luna.MyBio.PDB.PDBParser import PDBParser
from luna.mol.validator import MolValidator
from luna.mol.standardiser import ResiduesStandardiser
from luna.wrappers.base import MolWrapper
from luna.wrappers.obabel import convert_molecule
from luna.wrappers.rdkit import read_mol_from_file

from luna.util.exceptions import (IllegalArgumentError, MoleculeNotFoundError, ChainNotFoundError,
                                  FileNotCreated, PDBNotReadError, MoleculeSizeError, MoleculeObjectError)


logger = logging.getLogger()


ENTITY_LEVEL_NAME = {"A": "Atom",
                     "R": "Residue",
                     "C": "Chain",
                     "M": "Model",
                     "S": "Structure"}


def download_pdb(pdb_id, output_path=".", overwrite=False):
    """Download a PDB file from RCSB.org.

    Parameters
    ----------
    pdb_id : str
        4-symbols structure Id from PDB (e.g. 3J92).
    output_path : str
        Put the PDB file in this directory.
    overwrite : bool
        If True, overwrite any existing PDB files.
    """
    logger.debug("It will try to download the PDB '%s' and store it at the directory '%s'." % (pdb_id, output_path))

    if pdb_id is not None and pdb_id.strip() != "":
        if is_directory_valid(output_path):
            new_pdb_file = "%s/%s.pdb" % (output_path, pdb_id)
            # If the file already exists and is not to overwrite the file, do nothing.
            if overwrite is False and not exists(new_pdb_file):
                pdbl = PDBList()
                pdbl.retrieve_pdb_file(pdb_id, pdir=output_path, file_format="pdb", overwrite=overwrite)
                logger.debug("Download of the PDB '%s' completed." % pdb_id)

                # Rename files.
                cur_pdb_file = '%s/pdb%s.ent' % (output_path, pdb_id.lower())
                rename_pdb_file(cur_pdb_file, new_pdb_file)
            else:
                logger.debug("File '%s' already exists. It will not be downloaded again." % new_pdb_file)
    else:
        raise IllegalArgumentError("An empty PDB id ('%s') was informed." % pdb_id)


def parse_from_file(pdb_id, file):
    """Read a PDB file and return a :class:`~luna.MyBio.PDB.Structure.Structure` object.

    Parameters
    ----------
    pdb_id : str
        The structure identifier.
    file : str
        Pathname of the PDB file.

    Returns
    -------
    structure : :class:`~luna.MyBio.PDB.Structure.Structure`
        The parsed PDB file as a :class:`~luna.MyBio.PDB.Structure.Structure` object.

    Raises
    ------
    PDBNotReadError
        If the PDB file could not be parsed.
    """
    try:
        parser = PDBParser(PERMISSIVE=True, QUIET=True, FIX_EMPTY_CHAINS=True,
                           FIX_ATOM_NAME_CONFLICT=True, FIX_OBABEL_FLAGS=False)
        structure = parser.get_structure(pdb_id, file)
        return structure
    except Exception as e:
        logger.exception(e)
        raise PDBNotReadError("File '%s' could not be parsed." % file)


def save_to_file(entity, output_file, select=Select(), write_conects=True, write_end=True,
                 preserve_atom_numbering=True):
    """Write a Structure object (or a subset of a :class:`~luna.MyBio.PDB.Structure.Structure` object)
    into a file.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        The PDB object to be saved.
    output_file : str
        Save the selected atoms to this file.
    select : :class:`~luna.MyBio.PDB.PDBIO.Select`
        Decide which atoms will be saved at the PDB output. By default, all atoms are accepted.
    write_conects : bool
        If True, write CONECT records.
    write_end : bool
        If True, write the END record.
    preserve_atom_numbering : bool
        If True, preserve the atom numbering. Otherwise, re-enumerate the atom serial numbers.
    """
    try:
        io = PDBIO()
        io.set_structure(entity)
        io.save(output_file, select=select, write_conects=write_conects, write_end=write_end,
                preserve_atom_numbering=preserve_atom_numbering)
    except Exception as e:
        logger.exception(e)
        raise FileNotCreated("PDB file '%s' could not be created." % output_file)


def entity_to_string(entity, select=Select(), write_conects=True, write_end=True, preserve_atom_numbering=True):
    """Convert a Structure object (or a subset of a :class:`~luna.MyBio.PDB.Structure.Structure` object) to string.

    This function works on a structural level. That means if ``entity`` is not a :class:`~luna.MyBio.PDB.Structure.Structure` object,
    the structure will be recovered directly from ``entity``.
    Therefore, use ``select`` to select specific chains, residues, and atoms from the structure object.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        The PDB object to be converted.
    select : :class:`~luna.MyBio.PDB.PDBIO.Select`
        Decides which atoms will be saved at the output. By default, all atoms are accepted.
    write_conects : bool
        If True, writes CONECT records.
    write_end : bool
        If True, writes the END record.
    preserve_atom_numbering : bool
        If True, preserve the atom numbering. Otherwise, re-enumerate the atom serial numbers.

    Returns
    -------
    str
        The converted ``entity``.
    """
    fh = StringIO()
    io = PDBIO()
    io.set_structure(entity)
    io.save(fh, select=select, write_conects=write_conects, write_end=write_end, preserve_atom_numbering=preserve_atom_numbering)
    fh.seek(0)
    return ''.join(fh.readlines())


def get_entity_from_entry(entity, entry, model=0):
    """Get a :class:`~luna.MyBio.PDB.Entity.Chain` or :class:`~luna.MyBio.PDB.Residue.Residue`
    instance based on the provided entry ``entry``.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        The PDB object to recover the target entry from.
    entry : :class:`~luna.entry.Entry`
        The target entry.
    model : int
        The PDB model where the entry could be found. The default value is 0.

    Returns
    -------
    :class:`~luna.MyBio.PDB.Entity.Entity`
        The target entry.

    Raises
    ------
    MoleculeNotFoundError
        If the entry's molecule was not found in ``entity``.
    ChainNotFoundError
        If the entry's chain was not found in ``entity``.
    """
    structure = entity.get_parent_by_level("S")
    model = structure[model]

    if entry.chain_id in model.child_dict:
        chain = model[entry.chain_id]

        if entry.comp_name is not None and entry.comp_num is not None:
            ligand_key = entry.get_biopython_key()

            if ligand_key in chain.child_dict:
                target_entity = chain[ligand_key]
            else:
                raise MoleculeNotFoundError("Ligand '%s' does not exist in the PDB '%s'."
                                            % (entry.to_string(ENTRY_SEPARATOR), structure.get_id()))
        else:
            target_entity = chain
    else:
        raise ChainNotFoundError("The informed chain id '%s' for the ligand entry '%s' does not exist in the PDB '%s'." %
                                 (entry.chain_id, entry.to_string(ENTRY_SEPARATOR), structure.get_id()))

    return target_entity


def is_covalently_bound(atm1, atm2):
    """Verifies if atoms ``atm1`` and ``atm2`` are covalently bound."""

    # Distance atom-atom
    dist = atm1 - atm2
    # Covalent radius
    cov1 = ob.GetCovalentRad(ob.GetAtomicNum(atm1.element))
    cov2 = ob.GetCovalentRad(ob.GetAtomicNum(atm2.element))

    # OpenBabel thresholds.
    if 0.4 <= dist <= cov1 + cov2 + 0.45:
        return True
    return False


def get_residue_cov_bonds(residue, select=Select()):
    """Get covalently bound atoms from residues or other molecules.

    Parameters
    ----------
    residue : :class:`~luna.MyBio.PDB.Residue.Residue`
        The residue or other molecule from which covalently bound atoms will be recovered.
    select : :class:`~luna.MyBio.PDB.PDBIO.Select`
        Decides which atoms will be consired. By default, all atoms are accepted.

    Returns
    -------
    list of tuple(:class:`~luna.MyBio.PDB.Atom.Atom`, :class:`~luna.MyBio.PDB.Atom.Atom`)
        List of pairs of covalently bound atoms.
    """

    # Get valid atoms according to the provided selection function.
    trgt_res_atms = {atm.name: atm for atm in residue.get_atoms() if select.accept_atom(atm)}

    cov_bonds = []
    for atm1, atm2 in combinations(trgt_res_atms.values(), 2):
        if is_covalently_bound(atm1, atm2):
            cov_bonds.append((atm1, atm2))

    return cov_bonds


def get_residue_neighbors(residue, select=Select()):
    """Get all neighbors from a residue.

    In the case of an amino acid that is part of a peptide bond, its neighbors are any predecessor or successor residues.
    The same idea applies to nucleic acids.

    Parameters
    ----------
    residue : :class:`~luna.MyBio.PDB.Residue.Residue`
        The residue or other molecule from which covalently bound molecules will be recovered.
    select : :class:`~luna.MyBio.PDB.PDBIO.Select`
        Decides which atoms will be consired. By default, all atoms are accepted.

    Returns
    -------
    neighbors : dict of {str : :class:`~luna.MyBio.PDB.Residue.Residue`}
        A dictionary containing the predecessor (``previous``) or successor (``next``) molecules.
    """

    # Get valid atoms according to the provided selection function.
    trgt_res_atms = {atm.name: atm for atm in residue.get_atoms() if select.accept_atom(atm)}

    if residue.is_residue():
        if "N" not in trgt_res_atms:
            logger.debug("There is a missing N in the residue %s. It may have been filtered out by the provided "
                         "selection function. So, the predecessor residue cannot be identified." % residue)
        if "C" not in trgt_res_atms:
            logger.debug("There is a missing C in the residue %s. It may have been filtered out by the provided "
                         "selection function. So, the successor residue cannot be identified." % residue)

        # If neither N and C are available in the valid atom list.
        if "N" not in trgt_res_atms and "C" not in trgt_res_atms:
            return {}

        neighbors = {}
        # If the chain has a residue coming before the target residue.
        if residue.idx - 1 >= 0:
            # First residue before the target in the chain list.
            prev_res = residue.parent.child_list[residue.idx - 1]
            # Get valid atoms according to the provided selection function.
            prev_res_atms = {atm.name: atm for atm in prev_res.get_atoms() if select.accept_atom(atm)}

            # A peptide bond exists between the C of one amino acid and the N of another.
            if "C" in prev_res_atms and "N" in trgt_res_atms:
                if is_covalently_bound(trgt_res_atms["N"], prev_res_atms["C"]):
                    neighbors["previous"] = prev_res
                else:
                    logger.debug("The first residue before %s is too distant to fulfill the covalent thresholds. "
                                 "It may be an indication of bad atom positioning or that there are missing residues." % residue)
            elif "C" not in prev_res_atms:
                logger.debug("There is a missing C in %s, the first residue before %s in the chain list. "
                             "It may have been filtered out by the provided selection function. "
                             "So, the predecessor residue cannot be identified." % (prev_res, residue))
        # Otherwise, it could mean that the residue is the first one in the sequence or there are missing residues.
        else:
            logger.debug("The residue %s seems not to have any predecessor residue. "
                         "It may be the first in the chain sequence or there are missing residues." % residue)

        # If the chain has a residue coming after the target residue.
        if residue.idx + 1 < len(residue.parent.child_list):
            # First residue after the target in the chain list.
            next_res = residue.parent.child_list[residue.idx + 1]
            # Get valid atoms according to the provided selection function.
            next_res_atms = {atm.name: atm for atm in next_res.get_atoms() if select.accept_atom(atm)}

            # A peptide bond exists between the C of one amino acid and the N of another.
            if "C" in trgt_res_atms and "N" in next_res_atms:
                if is_covalently_bound(trgt_res_atms["C"], next_res_atms["N"]):
                    neighbors["next"] = next_res
                else:
                    logger.debug("The first residue after %s is too distant to fulfill the covalent thresholds. "
                                 "It may be an indication of bad atom positioning or that there are missing residues." % residue)
            elif "N" in next_res_atms:
                logger.debug("There is a missing N in %s, the first residue after %s in the chain list. "
                             "It may have been filtered out by the provided selection function. "
                             "So, the predecessor residue cannot be identified." % (next_res, residue))
        # Otherwise, it could mean that the residue is the last one in the sequence or there are missing residues.
        else:
            logger.debug("The residue %s seems not to have any successor residue. "
                         "It may be the last in the chain sequence or there are missing residues." % residue)
        return neighbors
    else:
        # First residue before the target in the chain list.
        prev_res = residue.parent.child_list[residue.idx - 1]
        # Get valid atoms according to the provided selection function.
        prev_res_atms = {atm.name: atm for atm in prev_res.get_atoms() if select.accept_atom(atm)}

        neighbors = {}
        # If the chain has a residue coming before the target residue.
        if residue.idx - 1 >= 0:
            # First residue before the target in the chain list.
            prev_res = residue.parent.child_list[residue.idx - 1]
            # Get valid atoms according to the provided selection function.
            prev_res_atms = {atm.name: atm for atm in prev_res.get_atoms() if select.accept_atom(atm)}

            for trgt_atm, prev_atm in product(trgt_res_atms.values(), prev_res_atms.values()):
                if is_covalently_bound(trgt_atm, prev_atm):
                    neighbors["previous"] = prev_res
                    break

            if "previous" not in neighbors:
                logger.debug("The first residue after %s is too distant to fulfill the covalent thresholds. "
                             "It may be an indication of bad atom positioning or that there are missing residues." % residue)
        # Otherwise, it could mean that the residue is the first one in the sequence or there are missing residues.
        else:
            logger.debug("The residue %s seems not to have any predecessor residue. "
                         "It may be the first in the chain sequence or there are missing residues." % residue)

        # If the chain has a residue coming after the target residue.
        if residue.idx + 1 < len(residue.parent.child_list):
            # First residue after the target in the chain list.
            next_res = residue.parent.child_list[residue.idx + 1]
            # Get valid atoms according to the provided selection function.
            next_res_atms = {atm.name: atm for atm in next_res.get_atoms() if select.accept_atom(atm)}

            # Check each pair of atoms for covalently bonded atoms.
            for trgt_atm, next_atm in product(trgt_res_atms.values(), next_res_atms.values()):
                if is_covalently_bound(trgt_atm, next_atm):
                    neighbors["next"] = next_res
                    break

            if "next" not in neighbors:
                logger.debug("The first residue after %s is too distant to fulfill the covalent thresholds. "
                             "It may be an indication of bad atom positioning or that there are missing residues." % residue)
        # Otherwise, it could mean that the residue is the last one in the sequence or there are missing residues.
        else:
            logger.debug("The residue %s seems not to have any successor residue. "
                         "It may be the last in the chain sequence or there are missing residues." % residue)
        return neighbors
    return {}


def biopython_entity_to_mol(entity, select=Select(), validate_mol=True, standardize_mol=True,
                            template=None, add_h=False, ph=None,
                            break_metal_bonds=False, mol_obj_type="rdkit", wrapped=True,
                            openbabel=OPENBABEL, tmp_path=None, keep_tmp_files=False):
    """Convert an object :class:`~luna.MyBio.PDB.Entity.Entity` to a
    molecular object (:class:`~luna.wrappers.base.MolWrapper`,
    :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`).

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Entity`
        The entity to be converted.
    select : :class:`~luna.MyBio.PDB.PDBIO.Select`
        Decide which atoms will be consired. By default, all atoms are accepted.
    validate_mol : bool
        If True, validate the converted molecule with :class:`~luna.mol.validator.MolValidator`.
    standardize_mol : bool
        If True, standardize the converted molecule. Currently, only residues are standardized as it
        uses :class:`luna.mol.standardiser.ResiduesStandardiser` by default.
    template : :class:`luna.mol.template.Template`, optional
        Fix the converted molecule's bond order based on the bond orders in a template molecule.
        The ``template`` should implement assign_bond_order().
    add_h : bool
        If True, add hydrogen to the converted molecule.
    ph : float, optional
        Add hydrogens considering pH ``ph``.
    break_metal_bonds : bool
        If True, remove covalent bonds between residues and metals and fix the residue bond order.
    mol_obj_type : {"rdkit", "openbabel"}
        If "rdkit", parse the converted molecule with RDKit and return an instance of :class:`rdkit.Chem.rdchem.Mol`.
        If "openbabel", parse the converted molecule with Open Babel and return an instance of
        :class:`openbabel.pybel.Molecule`.
        If ``wrapped`` is True, the molecular object will be wrapped with :class:`~luna.wrappers.base.MolWrapper`.
    wrapped : bool
        If True, wrap the molecular object with :class:`~luna.wrappers.base.MolWrapper`.
    openbabel : str
        Pathname to Open Babel. Even if ``mol_obj_type`` is set to 'rdkit', this function uses
        Open Babel to properly parse PDB files and to add hydrogens to the molecule
        considering different pHs.
    tmp_path : str
        A temporary directory to where temporary files will be saved.
        If not provided, the system's default temporary directory will be used instead.
    keep_tmp_files : bool
        If True, keep all temporary files. Otherwise, removes them in the end.

    Returns
    -------
    mol_obj : :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
        The converted molecule.
    ignored_atoms : list
        List of ignored atoms. Currently, ignored atoms may contain only metals.
    """

    tmp_path = tmp_path or tempfile.gettempdir()

    logger.debug("It will try to create a new MOL object from the provided entity.")
    logger.debug("Temporary files will be saved at '%s'." % tmp_path)

    # First it saves the selection into a PDB file and then it converts the file to .mol.
    # I had to do it because the OpenBabel 2.4.1 had a problem with some molecules containing aromatic rings.
    # In such cases, the aromatic ring was being wrongly perceived and some atoms received more double bonds than it
    # was expected. The version 2.3.2 works better. Therefore, I defined this version manually (openbabel property).
    filename = new_unique_filename(tmp_path)
    pdb_file = '%s_pdb-file.pdb' % filename

    logger.debug("First: try to create a new PDB file (%s) from the provided entity." % pdb_file)
    # Apparently, Open Babel creates a bug when it tries to parse a file with CONECTS containing serial numbers with more than 4 digits.
    # E.g.: 1OZH:A:HE3:1406, line CONECT162811627916282. By setting preserve_atom_numbering to False, it solves the problem.
    save_to_file(entity, pdb_file, select, preserve_atom_numbering=False)

    ini_input_file = pdb_file
    if template is not None:
        if entity.level == "R" and entity.is_hetatm():
            # Note that the template molecule should have no explicit hydrogens else the algorithm will fail.
            rdmol = read_mol_from_file(pdb_file, mol_format="pdb", removeHs=True)
            new_rdmol = template.assign_bond_order(rdmol, entity.resname)

            ini_input_file = '%s_tmp-mol-file.mol' % filename
            MolToMolFile(new_rdmol, ini_input_file)

            if not keep_tmp_files:
                remove_files([pdb_file])
        else:
            logger.warning("It cannot apply a template on the provided entity because it should be a single compound (Residue class).")

    # Convert the PDB file to Mol file with the proper protonation and hydrogen addition if required.
    mol_file = '%s_mol-file.mol' % filename
    ob_opt = {"error-level": 5}
    logger.debug("Next: try to convert the PDB file to .mol using Open Babel.")
    if add_h:
        logger.debug("Hydrogens will be added to the molecule.")
        if ph is not None:
            ob_opt["p"] = ph
        else:
            ob_opt["h"] = ""
    convert_molecule(ini_input_file, mol_file, opts=ob_opt, openbabel=openbabel)

    # Currently, ignored atoms are only metals.
    ignored_atoms = []

    mol_obj = None
    if validate_mol or standardize_mol:
        logger.debug("Now: a validation will be performed and it will try to fix some errors.")

        try:
            mol_obj = next(readfile("mol", mol_file))
        except Exception:
            raise MoleculeObjectError("An error occurred while parsing the file '%s' with Open Babel and the molecule "
                                      "object could not be created. Check the logs for more information." % mol_file)

        if standardize_mol:
            # Standardize a specific set of atoms if provided, otherwise use all atoms from entity.
            if isinstance(select, AtomSelector):
                target_atoms = select.entries
            else:
                target_atoms = Selection.unfold_entities(entity, 'A')

            # If the add_h property is set to False, the code will not remove any existing hydrogens from the PDB structure.
            # In these situations, the list of atoms may contain hydrogens. But, we do not need to attribute properties to hydrogens.
            # We just need them to correctly set properties to heavy atoms. So let's just ignore them.
            target_atoms = [atm for atm in target_atoms if select.accept_atom(atm) and atm.element != "H"]

            mol_obj = MolWrapper(mol_obj)
            if mol_obj.get_num_heavy_atoms() != len(target_atoms):
                raise MoleculeSizeError("The number of heavy atoms in the PDB selection and in the MOL file are different.")

            # Ignore hydrogen atoms.
            atm_obj_list = [atm for atm in mol_obj.get_atoms() if atm.get_atomic_num() != 1]

            atom_pairs = []
            for i, atm_obj in enumerate(atm_obj_list):
                if target_atoms[i].parent.is_residue():
                    atom_pairs.append((atm_obj, target_atoms[i]))

            rs = ResiduesStandardiser(break_metal_bonds=break_metal_bonds)
            mol_obj.unwrap().DeleteHydrogens()
            rs.standardise(atom_pairs)

            for i, atm_obj in enumerate(atm_obj_list):
                if atm_obj.get_id() in rs.removed_atoms:
                    ignored_atoms.append(target_atoms[i])

            # After standardizing residues, we need to recreate the Mol file, otherwise implicit hydrogens will not be included in the
            # MOL object and, therefore, their coordinates could not be accessed. If you try to generate coordinates directly from the
            # object, hydrogens will be incorrectly placed.
            mol_obj = PybelWrapper(mol_obj.unwrap())
            new_mol_file = '%s_tmp-mol-file.mol' % filename
            mol_obj.write("mol", new_mol_file, overwrite=True)
            # Overwrite mol_file by converting the new molecular file using the user specified parameters.
            # Note that right now it will add explicit hydrogens to the molecules according to the provided pH.
            convert_molecule(new_mol_file, mol_file, opts=ob_opt, openbabel=OPENBABEL)

            # Let's finally read the correct and standardized molecular file.
            try:
                mol_obj = next(readfile("mol", mol_file))
            except Exception:
                raise MoleculeObjectError("An error occurred while parsing the file '%s' with Open Babel and the molecule "
                                          "object could not be created. Check the logs for more information." % mol_file)

            # Remove temporary files.
            if not keep_tmp_files:
                remove_files([new_mol_file])

        mv = MolValidator()
        is_valid = mv.validate_mol(mol_obj)
        logger.debug('Validation finished!!!')

        if not is_valid:
            logger.warning("The molecular file '%s' contain invalid atoms. Check the logs for more information." % mol_file)

        if mol_obj_type == "rdkit":
            try:
                # The sanitization is set off. We will apply it in the next step.
                mol_obj = MolFromMolBlock(mol_obj.write('mol'), sanitize=False, removeHs=False)
                # Sanitize molecule is applied now, so we will be able to catch the exceptions raised by RDKit,
                # otherwise it would not be possible.
                SanitizeMol(mol_obj, SanitizeFlags.SANITIZE_ALL)
            except Exception:
                raise MoleculeObjectError("An error occurred while parsing the molecular block with RDKit. The block was "
                                          "generated by Open Babel from the file '%s'. Check the logs for more information." % mol_file)
    else:
        try:
            # Create a new Mol object.
            if mol_obj_type == "openbabel":
                mol_obj = next(readfile("mol", mol_file))
            else:
                # The sanitization is set off. We will apply it in the next statement.
                mol_obj = MolFromMolFile(mol_file, sanitize=False, removeHs=False)
                # Sanitize molecule is applied now, so we will be able to catch the exceptions raised by RDKit,
                # otherwise it would not be possible.
                SanitizeMol(mol_obj, SanitizeFlags.SANITIZE_ALL)
        except Exception:
            tool = "Open Babel" if mol_obj_type == "openbabel" else "RDKit"
            raise MoleculeObjectError("An error occurred while parsing the file '%s' with %s and the molecule "
                                      "object could not be created. Check the logs for more information." % (mol_file, tool))

    # Remove temporary files.
    if not keep_tmp_files:
        remove_files([ini_input_file, mol_file])

    if wrapped:
        mol_obj = MolWrapper(mol_obj)

    return mol_obj, ignored_atoms



import unittest

import sys
from os.path import dirname, abspath

sys.path.append(dirname(dirname(abspath(__file__))))

from luna.MyBio.PDB.Model import Model
from luna.MyBio.PDB.Chain import Chain
from luna.MyBio.extractor import Extractor


class ExtractorTest(unittest.TestCase):

    def test_init(self):
        # Wrong type.
        self.assertRaises(TypeError, Extractor, "3QL8")
        self.assertRaises(TypeError, Extractor, 1)
        self.assertRaises(TypeError, Extractor, int)
        self.assertRaises(TypeError, Extractor, self)
        self.assertRaises(TypeError, Extractor, Chain)
        self.assertRaises(TypeError, Extractor, None)

        # Wrong initialization.
        with self.assertRaises(TypeError):
            Extractor(Chain())
        # Wrong initialization.
        with self.assertRaises(TypeError):
            Extractor(Model())

        # Extractor correctly initialized.
        model = Model(0)
        extractor = Extractor(model)
        self.assertEqual(extractor.entity.id, model.id)
        self.assertNotEqual(extractor.entity.id, "A")
        self.assertNotEqual(extractor.entity.id, 1)

        extractor.extract_chains(["A"], "test.pdb")

        # Extractor correctly initialized.
        chain = Chain("A")
        extractor = Extractor(chain)
        self.assertEqual(extractor.entity.id, chain.id)
        self.assertNotEqual(extractor.entity.id, "B")
        self.assertNotEqual(extractor.entity.id, 1)


if __name__ == '__main__':
    unittest.main()



import re
import logging

from rdkit import DataStructs
from rdkit.ML.Cluster import Butina

from luna.util.exceptions import IllegalArgumentError

logger = logging.getLogger()


def available_similarity_functions():
    """Return a list of all similarity metrics available at RDKit."""
    regex = re.compile("Bulk([a-zA-Z]+)Similarity", flags=0)
    return list(filter(regex.match, dir(DataStructs)))


def calc_distance_matrix(fps, similarity_func="BulkTanimotoSimilarity"):
    """Calculate the pairwise distance (dissimilarity) between fingerprints in ``fps`` using
        the similarity metric ``similarity_func``.

    Parameters
    ----------
    fps : iterable of RDKit :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` or :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`
        A sequence of fingerprints.
    similarity_func : str
        A similarity metric to calculate the distance between the provided fingerprints. The default value is 'BulkTanimotoSimilarity'.

        To check out the list of available similarity metrics, call the function :py:meth:`available_similarity_functions`.

    Examples
    --------

    First, let's define a set of molecules.

    >>> from luna.wrappers.base import MolWrapper
    >>> mols = [MolWrapper.from_smiles("CCCCCC").unwrap(),
    ...         MolWrapper.from_smiles("CCCCCCCC").unwrap(),
    ...         MolWrapper.from_smiles("CCCCCCCCO").unwrap()]

    Now, we generate fingerprints for those molecules.

    >>> from luna.mol.fingerprint import generate_fp_for_mols
    >>> fps = [d["fp"] for d in generate_fp_for_mols(mols, "morgan_fp")]

    Finally, calculate the distance between the molecules based on their fingerprints.

    >>> from luna.mol.clustering import calc_distance_matrix
    >>> print(calc_distance_matrix(fps))
    [0.125, 0.46153846153846156, 0.3846153846153846]

    Returns
    -------
    distances : list of float
        Flattened diagonal matrix.
    """
    funcs = available_similarity_functions()
    if similarity_func not in funcs:
        raise IllegalArgumentError("Similarity function not available.")

    dists = []
    for i in range(1, len(fps)):
        if (similarity_func == "BulkTverskySimilarity"):
            params = [fps[i], fps[:i], 0, 1]
        else:
            params = [fps[i], fps[:i]]

        sims = getattr(DataStructs, similarity_func)(*params)
        dists.extend([1 - x for x in sims])

    return dists


def cluster_fps(fps, cutoff=0.2, similarity_func="BulkTanimotoSimilarity"):
    """Clusterize molecules based on fingerprints using the Butina clustering algorithm.

    Parameters
    ----------
    fps : iterable of RDKit :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` or :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`
        A sequence of fingerprints.
    cutoff : float
        Elements within this range of each other are considered to be neighbors.
    similarity_func : str
        A similarity metric to calculate the distance between the provided fingerprints.  The default value is 'BulkTanimotoSimilarity'.

        To check out the list of available similarity metrics, call the function :py:meth:`available_similarity_functions`.

    Examples
    --------

    First, let's define a set of molecules.

    >>> from luna.wrappers.base import MolWrapper
    >>> mols = [MolWrapper.from_smiles("CCCCCC").unwrap(),
    ...         MolWrapper.from_smiles("CCCCCCCC").unwrap(),
    ...         MolWrapper.from_smiles("CCCCCCCCO").unwrap()]

    Now, we generate fingerprints for those molecules.

    >>> from luna.mol.fingerprint import generate_fp_for_mols
    >>> fps = [d["fp"] for d in generate_fp_for_mols(mols, "morgan_fp")]

    Finally, clusterize the molecules based on their fingerprints.

    >>> from luna.mol.clustering import cluster_fps
    >>> print(cluster_fps(fps, cutoff=0.2))
    ((1, 0), (2,))

    Returns
    -------
    clusters : tuple of tuples
        Each cluster is defined as a tuple of tuples, where the first element for each cluster is its centroid.
    """
    logger.debug("Trying to clusterize %d molecules." % len(fps))
    logger.debug("Defined cutoff: %.2f. Defined similarity function: %s." % (cutoff, similarity_func))

    # first generate the distance matrix.
    dists = calc_distance_matrix(fps, similarity_func)
    logger.debug("Distance matrix created.")

    # now cluster the data.
    cs = Butina.ClusterData(dists, len(fps), cutoff, isDistData=True)
    logger.debug("Number of cluster(s) created: %d." % len(cs))

    return cs



import configparser
import logging

logger = logging.getLogger()


class Config:
    """
    Parser for configuration files.

    Parameters
    ----------
    conf_file : str
        The pathname for the configuration file.

    Attributes
    ----------

    config : `configparser.ConfigParser`
        The parsed configuration file.
    """

    def __init__(self, conf_file):
        self.config = configparser.ConfigParser(allow_no_value=True)
        try:
            self.config.read(conf_file)
        except Exception as e:
            logger.exception(e)
            raise IOError('Configuration file %s not read.' % conf_file)

    def get_section_map(self, section):
        """
        Try to access the section ``section`` from the parsed configuration file.
        """
        section_map = {}
        options = self.config.options(section)
        for option in options:
            try:
                section_map[option] = self.config.get(section, option)
            except Exception as e:
                logger.exception(e)
                section_map[option] = None
        return section_map

    def __getattr__(self, attr):
        if hasattr(self.config, attr):
            return getattr(self.config, attr)
        else:
            raise AttributeError("The attribute '%s' does not exist in the class %s." % (attr, self.__class__.__name__))



# Copyright (C) 2002, Thomas Hamelryck (thamelry@binf.ku.dk)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.


###################################################################

# Modifications included by Alexandre Fassio (alexandrefassio@dcc.ufmg.br).
# Date: 19/02/2018.

# 1) Inherit inhouse modifications. Package: MyBio.

# Each line or block with modifications contain a MODBY tag.

###################################################################


"""The structure class, representing a macromolecular structure."""

from luna.MyBio.PDB.Entity import Entity


class Structure(Entity):
    """The Structure class contains a collection of Model instances."""

    # MODBY: Alexandre Fassio
    # New property: PDB file.
    def __init__(self, id, pdb_file=None):
        self.pdb_file = pdb_file
        self.level = "S"
        Entity.__init__(self, id)

    # Special methods

    def __repr__(self):
        return "<Structure id=%s>" % self.get_id()

    # Private methods

    def _sort(self, m1, m2):
        """Sort models.

        This sorting function sorts the Model instances in the Structure instance.
        The sorting is done based on the model id, which is a simple int that
        reflects the order of the models in the PDB file.

        Arguments:
         - m1, m2 - Model instances

        """
        return cmp(m1.get_id(), m2.get_id())

    # Public

    def get_models(self):
        for m in self:
            yield m

    def get_chains(self):
        for m in self.get_models():
            for c in m:
                yield c

    def get_residues(self):
        for c in self.get_chains():
            for r in c:
                yield r

    def get_atoms(self):
        for r in self.get_residues():
            for a in r:
                yield a



# Copyright (C) 2002, Thomas Hamelryck (thamelry@binf.ku.dk)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.


###################################################################

# Modifications included by Alexandre Fassio (alexandrefassio@dcc.ufmg.br).
# Date: 19/02/2018.

# 1) Inherit inhouse modifications. Package: MyBio.

# Each line or block with modifications contain a MODBY tag.

###################################################################



"""Superimpose two structures."""

from __future__ import print_function

import numpy

from Bio.SVDSuperimposer import SVDSuperimposer

# MODBY: Alexandre Fassio
# Inherit inhouse modifications. Package: MyBio.
from luna.MyBio.PDB.PDBExceptions import PDBException


class Superimposer(object):
    """Rotate/translate one set of atoms on top of another,
    thereby minimizing the RMSD.
    """

    def __init__(self):
        self.rotran = None
        self.rms = None

    def set_atoms(self, fixed, moving):
        """Put (translate/rotate) the atoms in fixed on the atoms in
        moving, in such a way that the RMSD is minimized.

        @param fixed: list of (fixed) atoms
        @param moving: list of (moving) atoms
        @type fixed,moving: [L{Atom}, L{Atom},...]
        """
        if not (len(fixed) == len(moving)):
            raise PDBException("Fixed and moving atom lists differ in size")
        l = len(fixed)
        fixed_coord = numpy.zeros((l, 3))
        moving_coord = numpy.zeros((l, 3))
        for i in range(0, len(fixed)):
            fixed_coord[i] = fixed[i].get_coord()
            moving_coord[i] = moving[i].get_coord()
        sup = SVDSuperimposer()
        sup.set(fixed_coord, moving_coord)
        sup.run()
        self.rms = sup.get_rms()
        self.rotran = sup.get_rotran()

    def apply(self, atom_list):
        """Rotate/translate a list of atoms."""
        if self.rotran is None:
            raise PDBException("No transformation has been calculated yet")
        rot, tran = self.rotran
        rot = rot.astype('f')
        tran = tran.astype('f')
        for atom in atom_list:
            atom.transform(rot, tran)



import unittest

import sys
from os.path import dirname, abspath

sys.path.append(dirname(dirname(abspath(__file__))))

from luna.util.progress_tracker import *


class ProgressResultTest(unittest.TestCase):

    def test_init(self):
        # Wrong type.
        self.assertRaises(TypeError, ProgressResult, "3QL8")
        self.assertRaises(TypeError, ProgressResult, 1)
        self.assertRaises(TypeError, ProgressResult, ProgressData(1, 20))

    def test_attribute_access(self):
        # Access 'inputs' when there is invalid objects in 'results'.
        with self.assertRaises(TypeError):
            pr = ProgressResult(["a", "b", "c"])
            pr.inputs

        # Access 'outputs' when there is invalid objects in 'results'.
        with self.assertRaises(TypeError):
            pr = ProgressResult(["a", "b", "c"])
            pr.outputs

        # Access 'errors' when there is invalid objects in 'results'.
        with self.assertRaises(TypeError):
            pr = ProgressResult(["a", "b", "c"])
            pr.errors

    def test_append(self):

        # Appending invalid data to 'results'
        with self.assertRaises(TypeError):
            pr = ProgressResult()
            pr.append(1)

        # Appending invalid data to 'results'
        with self.assertRaises(TypeError):
            pr = ProgressResult()
            pr.append("string")

        # Appending invalid data to 'results'
        with self.assertRaises(TypeError):
            pr = ProgressResult()
            pr.append(1)

        # Appending invalid data to 'results'
        with self.assertRaises(TypeError):
            pr = ProgressResult([ProgressData(1, 10), ProgressData(2, 11)])
            pr.append(1)

        # Check valid appending.
        pr = ProgressResult()
        pr.append(ProgressData(1, 20))
        self.assertEqual(pr.inputs, [1])

        # Check valid appending.
        pr = ProgressResult([ProgressData(1, 10), ProgressData(2, 11)])
        pr.append(ProgressData(3, 12))
        self.assertEqual(pr.inputs, [1, 2, 3])

        # Check outputs for valid appending.
        pr = ProgressResult()
        pr.append(ProgressData(3, 12, output_data="a"))
        self.assertEqual(pr.outputs, [(3, "a")])
        self.assertNotEqual(pr.outputs, [(3, "b")])

        # Check outputs for valid initialization.
        pr = ProgressResult([ProgressData(3, 12, output_data="a")])
        self.assertEqual(pr.outputs, [(3, "a")])
        self.assertNotEqual(pr.outputs, [(3, "b")])
        self.assertNotEqual(pr.outputs, [(1, "a")])


if __name__ == '__main__':
    unittest.main()



from luna.mol.entry import MolFileEntry
from luna.interaction.calc import InteractionsManager
from luna.wrappers.pymol import PymolWrapper, PymolSessionManager, mybio_to_pymol_selection
from luna.util.file import is_file_valid
from luna.util.exceptions import PymolSessionNotInitialized
from luna.MyBio.util import entity_to_string


class InteractionViewer(PymolSessionManager):

    """Class that inherits from :class:`~luna.wrappers.pymol.PymolSessionManager` and implements :meth:`set_view`
    to depict interactions in Pymol and save the view as a Pymol session.

    This class can be used to visualize multiple complexes into the same Pymol session, for instance, to compare
    binding modes. To do so, it is recommended that the protein structures are in the same coordinate system, i.e.,
    they should be aligned first. This can be achieved with :func:`luna.align.tmalign.align_2struct`
    or any other tool of your preference.

    Parameters
    ----------
    show_hydrop_surface : bool
        If True, highlight hydrophobic surfaces. The default value is False
    **kwargs : dict, optional
        Extra arguments to `InteractionViewer`. Refer to :class:`~luna.wrappers.pymol.PymolSessionManager`
        documentation for a list of all possible arguments.

    Examples
    --------

    In the below examples, we will assume a LUNA project object named ``proj_obj`` already exists.

    **Example 1)** In the first example, we will visualize all interactions identified in the first protein-ligand complex.
    To do so, we will provide a tuple containing an :class:`~luna.interaction.calc.InteractionsManager` from where
    the interactions will be recovered.

    First access the property ``interactions_mngrs`` to get an iterable of
    :class:`~luna.interaction.calc.InteractionsManager` objects and get the first one.

    >>> interactions_mngr = list(proj_obj.interactions_mngrs)[0]

    As `InteractionViewer` expects a list of tuples, we will define it now. The first item of the tuple
    is the :class:`~luna.mol.entry.Entry` instance, which represents a ligand. This can be obtained directly from
    the :class:`~luna.interaction.calc.InteractionsManager` object.
    The :class:`~luna.mol.entry.Entry` instance is necessary because the second item in the tuple (interactions)
    may be an iterable of :class:`~luna.interaction.type.InteractionType` from where such information
    cannot be recovered. Finally, the third item can be either a PDB file or a directory.
    In this example, we will use the PDB directory defined during the LUNA project initialization.

    >>> inter_tuples = [(interactions_mngr.entry, interactions_mngr, proj_obj.pdb_path)]

    Now, we create a new `InteractionViewer` object and call
    :meth:`~luna.wrappers.pymol.PymolSessionManager.new_session`, which will initialize the session, depict the interactions, and save
    the session to an output PSE file.

    >>> inter_view = InteractionViewer()
    >>> inter_view.new_session(inter_tuples, "output.pse")


    **Example 2)** In this example, we will create a new Pymol session where a given set of interactions will be shown.
    Let's say, for instance, we only want to visualize hydrogen bonds.

    To do so, instead of defining a tuple with an :class:`~luna.interaction.calc.InteractionsManager` instance, we will
    define a list of :class:`~luna.interaction.type.InteractionType` objects, which will contain only hydrogen bonds.

    Let's start with the selection of hydrogen bonds. Here, we will use the built-in method
    :meth:`luna.interaction.calc.InteractionsManager.filter_by_types`, which permits to filter interactions by type.

    >>> interactions_mngr = list(proj_obj.interactions_mngrs)[0]
    >>> hydrogen_bonds = interactions_mngr.filter_by_types(['Hydrogen bond'])

    Now, we just create the tuple as we did before and define the list of interactions we want to depict in the Pymol session.

    >>> inter_tuples = [(interactions_mngr.entry, hydrogen_bonds, proj_obj.pdb_path)]

    Finally, we create a new `InteractionViewer` object and call
    :meth:`InteractionViewer.new_session`, which will initialize the session, depict the interactions, and save
    the session to an output PSE file.

    >>> inter_view = InteractionViewer()
    >>> inter_view.new_session(inter_tuples, "output.pse")

    **Example 3)** In this final example, we will create a new Pymol session to visualize all complexes in a LUNA project.
    To do so, we create a list with one tuple for each complex:

    >>> inter_tuples = [(im.entry, im, proj_obj.pdb_path) for im in proj_obj.interactions_mngrs]

    Then, as we did before, just create a new `InteractionViewer` object and call :meth:`InteractionViewer.new_session`.

    >>> inter_view = InteractionViewer()
    >>> inter_view.new_session(inter_tuples, "output.pse")

    **Note:** it is recommended that all complexes have the same atomic coordinates to make the analysis and comparisons easier.
    If that's not the case, you may want to align the structures first. To do so, you can use :func:`luna.align.tmalign.align_2struct`
    or any other tool of your preference.
    """

    def __init__(self, show_hydrop_surface=False, **kwargs):
        self.show_hydrop_surface = show_hydrop_surface
        super().__init__(**kwargs)

    def set_view(self, inter_tuples):
        """Depict interactions into the current Pymol session.

        Parameters
        ----------
        inter_tuples : iterable of tuple
            Each tuple must contain three items: an :class:`~luna.mol.entry.Entry` instance,
            an iterable of :class:`~luna.interaction.type.InteractionType` or an :class:`~luna.interaction.calc.InteractionsManager`,
            and a PDB file or the directory where the PDB file is located.
        """

        if not isinstance(self.wrapper, PymolWrapper):
            raise PymolSessionNotInitialized("No session was initialized.")

        for target_entry, inter_data, pathname in inter_tuples:

            if is_file_valid(pathname):
                pdb_file = pathname
            else:
                pdb_file = "%s/%s.pdb" % (pathname, target_entry.pdb_id)

            main_grp = target_entry.to_string(sep="-").replace("'", "-")

            mol_block = (entity_to_string(target_entry.get_biopython_structure())
                         if isinstance(target_entry, MolFileEntry) else None)

            # Load PDB and extract hetatm.
            self.load_pdb(pdb_file, main_grp, mol_block)

            if isinstance(inter_data, InteractionsManager):
                interactions = inter_data.interactions
            else:
                interactions = inter_data

            # Add interactions and styles.
            interacting_residue_sels = self.set_interactions_view(interactions, main_grp)
            if interacting_residue_sels:
                self.wrapper.select(name="%s.inter_residues" % main_grp, selection=" or ".join(interacting_residue_sels))

            if self.show_hydrop_surface:
                self.wrapper.color([("white", "%s and !name PS*" % main_grp)])

                # It will display all hydrophobic groups if an AtomGroupsManager is available.
                atm_grp_mngr = interactions[0].src_grp.manager
                if atm_grp_mngr is not None:
                    interacting_atms = set()
                    for atm_grp in atm_grp_mngr.filter_by_types(["Hydrophobe", "Hydrophobic"], must_contain_all=False):

                        inters = [i for i in atm_grp.interactions if i.type == "Hydrophobic"]

                        for comp in atm_grp.compounds:
                            comp_sel = "%s and %s" % (main_grp, mybio_to_pymol_selection(comp))
                            self.wrapper.show([("sticks", comp_sel)])

                        if len(inters) > 0:
                            for inter in inters:
                                for atm in inter.src_grp.atoms + inter.trgt_grp.atoms:
                                    if atm not in interacting_atms:
                                        self.wrapper.color([("pink", "%s and %s" % (main_grp, mybio_to_pymol_selection(atm)))])

                                for atm in inter.src_interacting_atms + inter.trgt_interacting_atms:
                                    self.wrapper.color([("hotpink", "%s and %s" % (main_grp, mybio_to_pymol_selection(atm)))])
                                    interacting_atms.add(atm)

                        else:
                            for atm in atm_grp.atoms:
                                self.wrapper.color([("wheat", "%s and %s" % (main_grp, mybio_to_pymol_selection(atm)))])

                    self.wrapper.hide([("sticks", "%s and not hetatm" % main_grp)])
                    self.wrapper.show([("sticks", "%s.inter_residues" % main_grp)])
                    self.wrapper.color([("white", "%s and not hetatm and not %s.inter_residues and !name PS*" % (main_grp, main_grp))])

                # Otherwise, it will display only hydrophobic groups comprising the interacting groups.
                else:
                    interacting_atms = set()
                    for inter in interactions:
                        if inter.type != "Hydrophobic":
                            continue

                        for atm in inter.src_grp.atoms + inter.trgt_grp.atoms:
                            if atm not in interacting_atms:
                                self.wrapper.color([("pink", "%s and %s" % (main_grp, mybio_to_pymol_selection(atm)))])

                        for atm in inter.src_interacting_atms + inter.trgt_interacting_atms:
                            self.wrapper.color([("hotpink", "%s and %s" % (main_grp, mybio_to_pymol_selection(atm)))])
                            interacting_atms.add(atm)

        self.set_last_details_to_view()



'''
http://pymolwiki.org/index.php/cgo_arrow

(c) 2013 Thomas Holder, Schrodinger Inc.

License: BSD-2-Clause
'''

from pymol import cmd, cgo, CmdException


def cgo_arrow(atom1='pk1', atom2='pk2', radius=0.5, gap=0.0, hlength=-1, hradius=-1,
              color='blue red', name=''):
    '''
DESCRIPTION

    Create a CGO arrow between two picked atoms.

ARGUMENTS

    atom1 = string: single atom selection or list of 3 floats {default: pk1}

    atom2 = string: single atom selection or list of 3 floats {default: pk2}

    radius = float: arrow radius {default: 0.5}

    gap = float: gap between arrow tips and the two atoms {default: 0.0}

    hlength = float: length of head

    hradius = float: radius of head

    color = string: one or two color names {default: blue red}

    name = string: name of CGO object
    '''
    from chempy import cpv

    radius, gap = float(radius), float(gap)
    hlength, hradius = float(hlength), float(hradius)

    try:
        color1, color2 = color.split()
    except:
        color1 = color2 = color
    color1 = list(cmd.get_color_tuple(color1))
    color2 = list(cmd.get_color_tuple(color2))

    def get_coord(v):
        if not isinstance(v, str):
            return v
        if v.startswith('['):
            return cmd.safe_list_eval(v)
        return cmd.get_atom_coords(v)

    xyz1 = get_coord(atom1)
    xyz2 = get_coord(atom2)
    normal = cpv.normalize(cpv.sub(xyz1, xyz2))

    if hlength < 0:
        hlength = radius * 3.0
    if hradius < 0:
        hradius = hlength * 0.6

    if gap:
        diff = cpv.scale(normal, gap)
        xyz1 = cpv.sub(xyz1, diff)
        xyz2 = cpv.add(xyz2, diff)

    xyz3 = cpv.add(cpv.scale(normal, hlength), xyz2)

    obj = [cgo.CYLINDER] + xyz1 + xyz3 + [radius] + color1 + color2 + \
          [cgo.CONE] + xyz3 + xyz2 + [hradius, 0.0] + color2 + color2 + \
          [1.0, 0.0]

    if not name:
        name = cmd.get_unused_name('arrow')

    cmd.load_cgo(obj, name)

cmd.extend('cgo_arrow', cgo_arrow)



from rdkit.Chem import MolFromMol2File, MolFromPDBFile, MolFromMolFile, MolFromMolBlock, MolFromMol2Block, SanitizeFlags, SanitizeMol
from xopen import xopen

from luna.util.file import get_file_format
from luna.util.exceptions import IllegalArgumentError

import re
import logging

logger = logging.getLogger()

RDKIT_FORMATS = ("mol2", "mol", "mdl", "sdf", "sd", "pdb")

REGEX_MOL_FILE = re.compile(r' (V2000|V3000)$')


def read_mol_from_file(mol_file, mol_format, sanitize=True, removeHs=True):
    """Read a molecule from ``mol_file`` using RDKit.

    Parameters
    ----------
    mol_file : str
        The pathname of the molecular file to be read.
    mol_format : str
        The molecular format of ``mol_file``.
    sanitize : bool
        If True (the default), sanitize the molecule.
    removeHs : bool
        If True (the default), remove explict hydrogens from the molecule.

    Returns
    -------
     : :class:`rdkit.Chem.rdchem.Mol` or None
        The parsed molecule or None in case the sanitization process fails.

    Raises
    ------
    IllegalArgumentError
        If the provided molecular format is not accepted by RDKit.

    Examples
    --------

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.rdkit import read_mol_from_file
    >>> rdk_mol = read_mol_from_file(mol_file=f"{LUNA_PATH}/tutorial/inputs/ZINC000007786517.mol",
    ...                              mol_format="mol")
    >>> print(rdk_mol.GetProp("_Name"))
     ZINC000007786517
    """

    if mol_format == "mol2":
        # First it creates the molecule without applying the sanitization function.
        rdk_mol = MolFromMol2File(mol_file, sanitize=False, removeHs=removeHs)
    elif mol_format == "pdb":
        # First it creates the molecule without applying the sanitization function.
        rdk_mol = MolFromPDBFile(mol_file, sanitize=False, removeHs=removeHs)
    elif mol_format in RDKIT_FORMATS:
        # First it creates the molecule without applying the sanitization function.
        rdk_mol = MolFromMolFile(mol_file, sanitize=False, removeHs=removeHs)
    else:
        raise IllegalArgumentError("Invalid '%s' format. The accepted formats are: %s." % (mol_format, ",".join(RDKIT_FORMATS)))

    # Now it sanitizes the molecule if necessary.
    # We do it here in order to catch an Exception while performing the sanitization.
    if sanitize:
        try:
            SanitizeMol(rdk_mol, SanitizeFlags.SANITIZE_ALL)
        except Exception as e:
            logger.exception(e)
            return None
    return rdk_mol


def new_mol_from_block(block, mol_format, sanitize=True, removeHs=True):
    """Read a molecule from a string block using RDKit.

    Parameters
    ----------
    block : str
        The molecular string block.
    mol_format : str
        The molecular format of ``block``.
    sanitize : bool
        If True (the default), sanitize the molecule.
    removeHs : bool
        If True (the default), remove explict hydrogens from the molecule.

    Returns
    -------
     : :class:`rdkit.Chem.rdchem.Mol` or None
        The parsed molecule or None in case the sanitization process fails.

    Raises
    ------
    IllegalArgumentError
        If the provided molecular format is not accepted by RDKit.

    Examples
    --------
    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.rdkit import new_mol_from_block
    >>> with open(f"{LUNA_PATH}/tutorial/inputs/GLY.sdf", "r") as IN:
    ...           mol_block = IN.read()
    >>> rdk_mol = new_mol_from_block(block=mol_block, mol_format="sdf")
    >>> print(rdk_mol.GetProp("_Name"))
    GLY
    """

    if mol_format == "mol2":
        # First it creates the molecule without applying the sanitization function.
        rdk_mol = MolFromMol2Block(block, sanitize=False, removeHs=removeHs)
    elif mol_format in RDKIT_FORMATS:
        # First it creates the molecule without applying the sanitization function.
        rdk_mol = MolFromMolBlock(block, sanitize=False, removeHs=removeHs)
    else:
        raise IllegalArgumentError("Invalid '%s' format. The accepted formats are: %s." % (mol_format, ",".join(RDKIT_FORMATS)))

    # Now it sanitizes the molecule if necessary.
    # We do it here in order to catch an Exception while performing the sanitization.
    if sanitize:
        try:
            SanitizeMol(rdk_mol, SanitizeFlags.SANITIZE_ALL)
        except Exception as e:
            logger.exception(e)
            return None
    return rdk_mol


def read_multimol_file(mol_file, targets=None, mol_format=None, sanitize=True, removeHs=True):
    """Read molecules from a multimolecular file using RDKit.

    Parameters
    ----------
    mol_file : str
        The pathname of the molecular file to be read.
    targets : iterable of str
        Only parses molecules, given by their ids, defined in ``targets``.
    mol_format : str, optional
        The molecular format of ``mol_file``.
        If not provided, the format will be defined by the file extension.
    sanitize : bool
        If True (the default), sanitize the molecule.
    removeHs : bool
        If True (the default), remove explict hydrogens from the molecule.

    Yields
    -------
     : tuple of (:class:`rdkit.Chem.rdchem.Mol`, int)
        A tuple containing the parsed molecule and its id.

    Raises
    ------
    IllegalArgumentError
        If the provided or identified molecular format is not accepted by RDKit.

    Examples
    --------

    In this first example, we will read all molecules from a multimolecular file.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.rdkit import read_multimol_file
    >>> for mol_tuple in read_multimol_file(mol_file=f"{LUNA_PATH}/tutorial/inputs/ligands.mol2"):
    ...     print(mol_tuple[0].GetProp("_Name"))
     ZINC000343043015
     ZINC000065293174
     ZINC000575033470
     ZINC000096459890
     ZINC000012442563

    Now, we will only read two molecules (ZINC000065293174, ZINC000096459890) from the same multimolecular file.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.rdkit import read_multimol_file
    >>> for mol_tuple in read_multimol_file(mol_file=f"{LUNA_PATH}/tutorial/inputs/ligands.mol2",
    ...                                     targets=["ZINC000065293174", "ZINC000096459890"]):
    ...     print(mol_tuple[0].GetProp("_Name"))
     ZINC000065293174
     ZINC000096459890
    """

    def apply_mol_format(lines):
        # Check if the MOL file contains a valid header comprising a Title line, Program/file timestamp line, and a Comment line.
        # Thus, the Counts line must be in line 4 (index 3). If it's not, add blank lines to match the format.
        # This amendment is necessary otherwise RDKit will crash.
        if not REGEX_MOL_FILE.search(lines[3]):
            logger.warning("While parsing the file '%s', we found a molecule starting at line #%d that does not contain a valid header. "
                           "We will add empty lines to its header to match the MOL file format." % (mol_file, mol_starts_at))

            counts_line_pos = None
            for i, line in enumerate(lines):
                if REGEX_MOL_FILE.search(line):
                    counts_line_pos = i
                    break

            missing_lines = ["\n"] * (3 - counts_line_pos)
            lines = missing_lines + lines

        if lines[-1].strip() != "M  END":
            logger.warning("While parsing the file '%s', we found a molecule starting at line #%d that does not contain the END line. "
                           "We will add it to the block end to match the MOL file format." % (mol_file, mol_starts_at))
            lines.append("M  END\n")
        return lines

    ext = mol_format or get_file_format(mol_file, ignore_compression=True)

    if ext not in RDKIT_FORMATS:
        raise IllegalArgumentError("Format '%s' informed or assumed from the filename is invalid. The accepted formats are: %s."
                                   % (ext, ",".join(RDKIT_FORMATS)))

    with xopen(mol_file, "r") as IN:
        if targets is not None:
            targets = set(targets)

        mol = []
        ignore_lines = False
        line_count = 0
        while True:
            try:
                line = IN.readline()
                line_count += 1
                # readline() returns empty strings when EOF is reached.
                if not line:
                    raise StopIteration
                # Ignore new lines before a molecule block.
                if len(mol) == 0 and line == "\n":
                    continue
                # Save the line in which the new molecule starts.
                if len(mol) == 0:
                    mol_starts_at = line_count

                if ext == "mol2":
                    if line.startswith("#"):
                        continue
                    # New molecule block definition is starting...
                    if line.startswith("@<TRIPOS>MOLECULE"):
                        # New molecule identified but an old one already exists.
                        if mol:
                            mol_id = mol[1].strip()
                            # If a target list is not informed, create a new molecule object.
                            if targets is None:
                                # Create a new RDKit object
                                rdk_mol = new_mol_from_block("".join(mol), ext, sanitize, removeHs)
                                yield ((rdk_mol, mol_id))
                            # Otherwise, create a new molecule only if it is in the list.
                            elif mol_id in targets:
                                targets.remove(mol_id)
                                # Create a new RDKit object
                                rdk_mol = new_mol_from_block("".join(mol), ext, sanitize, removeHs)
                                yield ((rdk_mol, mol_id))
                        # Restart the molecule block.
                        mol = []
                    mol.append(line)
                else:
                    if line.startswith("M  END"):
                        mol.append(line)
                        # Fix the MOL block in cases where the header or end lines do not match the MOL file format.
                        mol = apply_mol_format(mol)
                        mol_id = mol[0].strip()

                        # If a target list is informed, create a new molecule only if it is in the list.
                        if targets is None:
                            # Create a new RDKit object
                            rdk_mol = new_mol_from_block("".join(mol), ext, sanitize, removeHs)
                            yield((rdk_mol, mol_id))

                        elif mol[0].strip() in targets:
                            targets.remove(mol_id)
                            # Create a new RDKit object
                            rdk_mol = new_mol_from_block("".join(mol), ext, sanitize, removeHs)
                            yield((rdk_mol, mol_id))
                        # Restart the molecule block.
                        mol = []
                        # After finding a molecule block, ignore any following lines until it finds a line "$$$$".
                        ignore_lines = True
                    elif line.startswith("$$$$") is False:
                        # Ignore lines util it finds a line "$$$$".
                        if ignore_lines:
                            continue
                        mol.append(line)
                    else:
                        ignore_lines = False

            except StopIteration:
                if mol:
                    if ext == "mol":
                        # Fix the MOL block in cases where the header or end lines do not match the MOL file format.
                        mol = apply_mol_format(mol)

                    mol_id = mol[1].strip() if ext == "mol2" else mol[0].strip()

                    # If a target list is informed, create a new molecule only if it is in the list.
                    if targets is None or (targets is not None and mol_id in targets):
                        # Create a new RDKit object
                        rdk_mol = new_mol_from_block("".join(mol), ext, sanitize, removeHs)
                        yield ((rdk_mol, mol_id))
                break

            # If all target compounds were already found, just break the loop.
            if targets is not None and len(targets) == 0:
                break



import unittest
from os.path import dirname, abspath
import sys

luna_path = dirname(dirname(abspath(__file__)))
sys.path.append(luna_path)


from luna.interaction.config import InteractionConfig, DefaultInteractionConfig


class InteractionConfigTest(unittest.TestCase):

    def test_init(self):

        # Wrong type.
        self.assertRaises(ValueError, InteractionConfig, "error")
        self.assertRaises(TypeError, InteractionConfig, 1)
        self.assertRaises(TypeError, InteractionConfig, {1, 2, 3})
        self.assertRaises(TypeError, InteractionConfig, [1, 2, 3])
        self.assertRaises(TypeError, DefaultInteractionConfig, "error")
        self.assertRaises(TypeError, DefaultInteractionConfig, 1)
        self.assertRaises(TypeError, DefaultInteractionConfig, {1, 2, 3})
        self.assertRaises(TypeError, DefaultInteractionConfig, [1, 2, 3])
        self.assertRaises(TypeError, DefaultInteractionConfig, {"param1": 2.5, "param2": 5})
        self.assertRaises(TypeError, DefaultInteractionConfig, None)

        # Correctly initialized.
        config = InteractionConfig()
        self.assertEqual(config, {})
        config = InteractionConfig(None)
        self.assertEqual(config, {})
        config = InteractionConfig([])
        self.assertEqual(config, {})

        # Correctly initialized.
        config = InteractionConfig([("param1", 2.5), ("param2", 5)])
        self.assertEqual(config["param1"], 2.5)
        self.assertEqual(config["param2"], 5)

        # Correctly initialized.
        config = InteractionConfig({"param1": 2.5, "param2": 5})
        self.assertEqual(config["param1"], 2.5)
        self.assertEqual(config["param2"], 5)

        # Correctly initialized.
        config = InteractionConfig()
        config["param1"] = 2.5
        config["param2"] = 2
        self.assertEqual(config["param1"], 2.5)
        self.assertEqual(config["param2"], 2)
        self.assertNotEqual(config["param2"], 5)

        # Correctly initialized.
        config = InteractionConfig({"param1": 2.5, "param2": 5})
        self.assertEqual(config.params, ['param1', 'param2'])

        # Correctly initialized.
        config = DefaultInteractionConfig()
        self.assertEqual(config.params, ['boundary_cutoff', 'max_an_ey_ang_ortho_multipolar_inter',
                                         'max_an_ey_ang_para_multipolar_inter', 'max_cc_dist_amide_pi_inter',
                                         'max_cc_dist_pi_pi_inter', 'max_da_dist_hb_inter', 'max_da_dist_whb_inter',
                                         'max_dc_dist_whb_inter', 'max_dihed_ang_amide_pi_inter',
                                         'max_dihed_ang_slope_pi_pi_inter', 'max_disp_ang_ion_multipole_inter',
                                         'max_disp_ang_multipolar_inter', 'max_disp_ang_offset_pi_pi_inter',
                                         'max_disp_ang_pi_pi_inter', 'max_disp_ang_whb_inter', 'max_disp_ang_xbond_inter',
                                         'max_disp_ang_ybond_inter', 'max_dist_attract_inter', 'max_dist_cation_pi_inter',
                                         'max_dist_hydrop_inter', 'max_dist_proximal', 'max_dist_repuls_inter',
                                         'max_ha_dist_hb_inter', 'max_ha_dist_whb_inter', 'max_hc_dist_whb_inter',
                                         'max_id_dist_ion_multipole_inter', 'max_ne_dist_multipolar_inter',
                                         'max_ney_ang_multipolar_inter', 'max_xa_dist_xbond_inter', 'max_xc_dist_xbond_inter',
                                         'max_ya_dist_ybond_inter', 'max_yc_dist_ybond_inter',
                                         'min_an_ey_ang_antipara_multipolar_inter', 'min_an_ey_ang_ortho_multipolar_inter',
                                         'min_bond_separation', 'min_bond_separation_for_clash', 'min_cxa_ang_xbond_inter',
                                         'min_dar_ang_hb_inter', 'min_dar_ang_whb_inter', 'min_dha_ang_hb_inter',
                                         'min_dha_ang_whb_inter', 'min_dhc_ang_whb_inter', 'min_dihed_ang_slope_pi_pi_inter',
                                         'min_disp_ang_offset_pi_pi_inter', 'min_dist_proximal', 'min_har_ang_hb_inter',
                                         'min_har_ang_whb_inter', 'min_idy_ang_ion_multipole_inter', 'min_inter_atom_in_surf',
                                         'min_ney_ang_multipolar_inter', 'min_rya_ang_ybond_inter', 'min_surf_size',
                                         'min_xar_ang_xbond_inter', 'min_yan_ang_ybond_inter', 'vdw_clash_tolerance',
                                         'vdw_tolerance'])
        config["boundary_cutoff"] = 5
        self.assertEqual(config["boundary_cutoff"], 5)

        # Access to nonexistent keys.
        config = DefaultInteractionConfig()
        self.assertRaises(KeyError, lambda: config["error"])

        # Access to removed keys.
        config = InteractionConfig({"param1": 2.5})
        del config["param1"]
        self.assertRaises(KeyError, lambda: config["param1"])

        # Access to nonexistent attributes.
        config = InteractionConfig({"param1": 2.5})
        self.assertRaises(AttributeError, lambda: config.param1)


if __name__ == '__main__':
    unittest.main()



from luna.MyBio.util import save_to_file
from luna.MyBio.selector import ChainSelector, ResidueSelector
from luna.MyBio.PDB.Model import Model
from luna.MyBio.PDB.Chain import Chain

import logging

logger = logging.getLogger()


class Extractor():
    """Extract chains or residues from ``entity``.

    Parameters
    ----------
    entity : :class:`~luna.MyBio.PDB.Entity.Model` or :class:`~luna.MyBio.PDB.Entity.Chain`
        A model or chain object from where chains and residues will be extracted, respectively.
    """

    def __init__(self, entity):
        self.entity = entity

    @property
    def entity(self):
        return self._entity

    @entity.setter
    def entity(self, entity):
        if not isinstance(entity, (Model, Chain)):
            raise TypeError("Only objects of type %s or %s are accepted." % (Model.__module__, Chain.__module__))
        self._entity = entity

    def extract_chains(self, chains, output_file):
        """Extract chains from ``entity`` and save it to ``output_file``.

        Parameters
        ----------
        chains : iterable of str
            A sequence of chains to extract from model ``entity``.
        output_file : str
            Save extracted chains to this file.
        """
        if self.entity.level != "M":
            logger.warning("Entity must be a Model object.")
        else:
            chain_sel = set()
            for chain in chains:
                if chain in self.entity.child_dict:
                    chain_sel.add(self.entity[chain])
                else:
                    logger.warning("Chain %s does not exist." % chain)

            if len(chain_sel) > 0:
                save_to_file(self.entity, output_file, ChainSelector(chain_sel))
            else:
                logger.warning("No valid chain to extract.")

    def extract_residues(self, residues, output_file):
        """Extract residues from ``entity`` and save it to ``output_file``.

        Parameters
        ----------
        residues : iterable
            A sequence of residues to extract from chain ``entity``.
        output_file : str
            Save extracted residues to to this file.
        """
        if self.entity.level != "C":
            logger.warning("Entity must be a Chain object.")
        else:
            res_sel = set()
            for res in residues:
                if res in self.entity.child_dict:
                    res_sel.add(self.entity[res])
                else:
                    logger.warning("Residue %s does not exist." % str(res))

            if len(res_sel) > 0:
                save_to_file(self.entity, output_file, ResidueSelector(res_sel))
            else:
                logger.warning("No valid residue to extract.")



import unittest

import sys
from os.path import dirname, abspath

sys.path.append(dirname(dirname(abspath(__file__))))

from luna.mol.entry import *
from luna.util.exceptions import InvalidEntry, IllegalArgumentError, MoleculeObjectError, MoleculeObjectTypeError, MoleculeNotFoundError


class EntryTest(unittest.TestCase):

    def test_init(self):
        # Missing obligatory parameter.
        self.assertRaises(TypeError, Entry, "3QL8")

        # Invalid chain id format.
        self.assertRaises(InvalidEntry, Entry, "3QL8", "AA")

        # Invalid chain id format.
        self.assertRaises(InvalidEntry, Entry, "3QL8", "+1")

        # Missing obligatory parameter when a compound information is provided.
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", "NAG")
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", comp_name="NAG")
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", comp_num=100)

        # Valid PDB and chain ID formats.
        self.assertEqual(Entry("3QL8", "A").to_string(), "3QL8:A")
        self.assertEqual(Entry("3QL8", "1").to_string(), "3QL8:1")
        self.assertEqual(Entry("3QL8", "A", None, None).to_string(), "3QL8:A")

        # Valid filenames.
        long_filename = "Q411SM6ZMPI6CBUKZSX10IA9M2923PUIHY88URUWC0R1ZOLQLJKXK9VATXRI3CM9W7RU409M19CG6S55TQ5L8L03HIKKL2HSEBTJ2VT7Y4G2JTUB558JOG2F3H9O524MMSYM0M3A347Q2JN77XG4MLAR3XI3EULR6WMVZ6U6NMASHPSUBE28ETOB0S1M0PTDX4KW45KLYC0JL5E77HKLAQK97NNOVB15B4SGVEVSZ90HZ3IMWZLF361AM0M0NSS"
        self.assertEqual(Entry(long_filename, "B").to_string(), "%s:B" % long_filename)
        self.assertEqual(Entry('file1.my_test=1a', "B").to_string(), "file1.my_test=1a:B")

        # Long filename.
        self.assertRaises(InvalidEntry, Entry, "4Q411SM6ZMPI6CBUKZSX10IA9M2923PUIHY88URUWC0R1ZOLQLJKXK9VATXRI3CM9W7RU409M19CG6S55TQ5L8L03HIKKL2HSEBTJ2VT7Y4G2JTUB558JOG2F3H9O524MMSYM0M3A347Q2JN77XG4MLAR3XI3EULR6WMVZ6U6NMASHPSUBE28ETOB0S1M0PTDX4KW45KLYC0JL5E77HKLAQK97NNOVB15B4SGVEVSZ90HZ3IMWZLF361AM0M0NSS", "B")

        # Valid compound entries.
        self.assertEqual(Entry("3QL8", "A", comp_name="X02", comp_num=100).to_string(), "3QL8:A:X02:100")
        self.assertEqual(Entry("3QL8", "A", "X02", 100).to_string(), "3QL8:A:X02:100")
        self.assertEqual(Entry("3QL8", "A", "X02", "100").to_string(), "3QL8:A:X02:100")
        self.assertEqual(Entry("3QL8", "A", 150, "100").to_string(), "3QL8:A:150:100")
        self.assertEqual(Entry("3QL8", "A", "X02", -100).to_string(), "3QL8:A:X02:-100")
        self.assertEqual(Entry("3QL8", "A", "X02", "+100").to_string(), "3QL8:A:X02:100")
        self.assertEqual(Entry("3NS9", "A", "NS9", "0").to_string(), "3NS9:A:NS9:0")

        # Valid compound entries with different compound name forms.
        self.assertEqual(Entry("3QL8", "A", "X", 533).to_string(), "3QL8:A:X:533")
        self.assertEqual(Entry("3QL8", "A", "XA", 533).to_string(), "3QL8:A:XA:533")
        self.assertEqual(Entry("3QL8", "A", "XAA", 533).to_string(), "3QL8:A:XAA:533")
        self.assertEqual(Entry("3QL8", "A", "Cl-", 533).to_string(), "3QL8:A:Cl-:533")
        self.assertEqual(Entry("3QL8", "A", "Na+", 533).to_string(), "3QL8:A:Na+:533")
        self.assertEqual(Entry("3QL8", "A", "NA+", 533).to_string(), "3QL8:A:NA+:533")
        self.assertEqual(Entry("3QL8", "A", "NA-", 533).to_string(), "3QL8:A:NA-:533")
        self.assertEqual(Entry("3QL8", "A", "N++", 533).to_string(), "3QL8:A:N++:533")
        self.assertEqual(Entry("3QL8", "A", "O--", 533).to_string(), "3QL8:A:O--:533")

        # Valid compound entries with Icode
        self.assertEqual(Entry("3QL8", "A", "X01", "100", "A").to_string(), "3QL8:A:X01:100A")

        # Invalid compound entries.
        self.assertRaises(InvalidEntry, Entry, "3QL8", "A", "X043", 10)
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", "X01", "XAA")
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", "X01", "X")
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", None, "X")
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", "X01", None)

        # Invalid compound entries with Icode
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", "X01", 100, "1")
        self.assertRaises(IllegalArgumentError, Entry, "3QL8", "A", "X01", 100, 1)

    def test_property_update(self):
        entry = Entry("3QL8", "A")

        with self.assertRaises(AttributeError):
            entry.pdb_id = "3QQK"

        with self.assertRaises(AttributeError):
            entry.chain_id = "C"

        with self.assertRaises(AttributeError):
            entry.comp_name = "X05"

        with self.assertRaises(AttributeError):
            entry.comp_num = 100

        with self.assertRaises(AttributeError):
            entry.comp_icode = "B"

        entry.is_hetatm = False
        self.assertEqual(entry.is_hetatm, False)

        entry.sep = "_"
        self.assertEqual(entry.to_string(), "3QL8_A")

    def test_from_string(self):
        # Missing obligatory parameter.
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8")

        # Invalid chain id format.
        self.assertRaises(InvalidEntry, Entry, "3QL8", "AA")

        # Invalid chain id format.
        self.assertRaises(InvalidEntry, Entry.from_string, "3QL8:+1")

        # Missing obligatory parameter when a compound information is provided.
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:NAG")
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:100")

        # Valid PDB and chain ID formats.
        self.assertEqual(Entry.from_string("3QL8:A").to_string(), "3QL8:A")
        self.assertEqual(Entry.from_string("3QL8:1").to_string(), "3QL8:1")
        self.assertEqual(Entry.from_string("3QL8-A", sep="-").to_string(), "3QL8-A")
        self.assertEqual(Entry.from_string("3QL8_A", sep="_").to_string(), "3QL8_A")

        # Valid filenames.
        long_filename = "Q411SM6ZMPI6CBUKZSX10IA9M2923PUIHY88URUWC0R1ZOLQLJKXK9VATXRI3CM9W7RU409M19CG6S55TQ5L8L03HIKKL2HSEBTJ2VT7Y4G2JTUB558JOG2F3H9O524MMSYM0M3A347Q2JN77XG4MLAR3XI3EULR6WMVZ6U6NMASHPSUBE28ETOB0S1M0PTDX4KW45KLYC0JL5E77HKLAQK97NNOVB15B4SGVEVSZ90HZ3IMWZLF361AM0M0NSS"
        self.assertEqual(Entry.from_string("%s:B" % long_filename).to_string(), "%s:B" % long_filename)
        self.assertEqual(Entry.from_string('file1.my_test=1a:B').to_string(), "file1.my_test=1a:B")

        # Long filename.
        self.assertRaises(InvalidEntry, Entry.from_string, "4Q411SM6ZMPI6CBUKZSX10IA9M2923PUIHY88URUWC0R1ZOLQLJKXK9VATXRI3CM9W7RU409M19CG6S55TQ5L8L03HIKKL2HSEBTJ2VT7Y4G2JTUB558JOG2F3H9O524MMSYM0M3A347Q2JN77XG4MLAR3XI3EULR6WMVZ6U6NMASHPSUBE28ETOB0S1M0PTDX4KW45KLYC0JL5E77HKLAQK97NNOVB15B4SGVEVSZ90HZ3IMWZLF361AM0M0NSS:B")

        # Valid compound entries.
        self.assertEqual(Entry.from_string("3QL8:A:X02:100").to_string(), "3QL8:A:X02:100")
        self.assertEqual(Entry.from_string("3QL8:A:150:100").to_string(), "3QL8:A:150:100")
        self.assertEqual(Entry.from_string("3QL8:A:X02:-100").to_string(), "3QL8:A:X02:-100")
        self.assertEqual(Entry.from_string("3QL8:A:X02:+100").to_string(), "3QL8:A:X02:100")

        # Valid compound entries with different compound name forms.
        self.assertEqual(Entry.from_string("3QL8:A:X:533").to_string(), "3QL8:A:X:533")
        self.assertEqual(Entry.from_string("3QL8:A:XA:533").to_string(), "3QL8:A:XA:533")
        self.assertEqual(Entry.from_string("3QL8:A:XAA:533").to_string(), "3QL8:A:XAA:533")
        self.assertEqual(Entry.from_string("3QL8:A:Cl-:533").to_string(), "3QL8:A:Cl-:533")
        self.assertEqual(Entry.from_string("3QL8:A:Na+:533").to_string(), "3QL8:A:Na+:533")
        self.assertEqual(Entry.from_string("3QL8:A:NA+:533").to_string(), "3QL8:A:NA+:533")
        self.assertEqual(Entry.from_string("3QL8:A:NA-:533").to_string(), "3QL8:A:NA-:533")
        self.assertEqual(Entry.from_string("3QL8:A:N++:533").to_string(), "3QL8:A:N++:533")
        self.assertEqual(Entry.from_string("3QL8:A:O--:533").to_string(), "3QL8:A:O--:533")
        self.assertEqual(Entry.from_string("3QL8:A:X02:9999").to_string(), "3QL8:A:X02:9999")

        # Valid compound entries with Icode
        self.assertEqual(Entry.from_string("3QL8:A:X01:100A").to_string(), "3QL8:A:X01:100A")

        # Invalid compound entries.
        self.assertRaises(InvalidEntry, Entry.from_string, "3QL8:A:X043:10")
        self.assertRaises(InvalidEntry, Entry.from_string, "3QL8:A:X02:99999")
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:X04:10:A")
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:X04:10:A:10")
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:X01:XAA")
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:X01:X")
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:None:X")
        self.assertRaises(IllegalArgumentError, Entry.from_string, "3QL8:A:X01")

    def test_get_biopython_key(self):
        entry = Entry("3QL8", "A")
        self.assertEqual(entry.get_biopython_key(), "A")

        entry = Entry("3QL8", "A", "X02", 104)
        self.assertEqual(entry.get_biopython_key(), ("H_X02", 104, " "))

        entry = Entry("3QL8", "A", "X02", 104, "A")
        self.assertEqual(entry.get_biopython_key(), ("H_X02", 104, "A"))

        entry = Entry("3QL8", "A", "X02", 104, " ")
        self.assertEqual(entry.get_biopython_key(), ("H_X02", 104, " "))

        entry = Entry("3QL8", "A", "X02", 104, None)
        self.assertEqual(entry.get_biopython_key(), ("H_X02", 104, " "))

        entry = Entry("3QL8", "A", "X02", 104, is_hetatm=False)
        self.assertEqual(entry.get_biopython_key(), (" ", 104, " "))

        entry = Entry("3QL8", "A", "HOH", 104, is_hetatm=False)
        self.assertEqual(entry.get_biopython_key(), ("W", 104, " "))

        entry = Entry("3QL8", "A", "HOH", 104, is_hetatm=True)
        self.assertEqual(entry.get_biopython_key(), ("W", 104, " "))

        entry = Entry("3QL8", "A", "WAT", 104, is_hetatm=False)
        self.assertEqual(entry.get_biopython_key(), ("W", 104, " "))


if __name__ == '__main__':
    unittest.main()



"""

LUNA: drug discovery toolkit.

If you have downloaded and uncompressed the LUNA source code, or fetched it from git,
for the simplest installation just type the command::

    python setup.py install

However, you would normally install the latest Biopython release from the PyPI archive
with::

    pip install luna

For more in-depth instructions, see the installation section of the
Biopython manual, linked to from:

http://luna-toolkit.readthedocs.io

Contributions, feature requests, and bug reports are greatly appreciated.
Please consult the issue tracker:

https://github.com/keiserlab/LUNA/issues

"""

from setuptools import setup, find_packages
from luna.version import version


__author__ = "Alexandre Fassio"
__date__ = "March 2022"
__maintainer__ = "Alexandre Fassio"
__email__ = "afassio@keiserlab.org"
__status__ = "Production"


def get_readme():
    with open('README.rst') as f:
        return f.read()


requirements = [
    'biopython==1.72',
    'colorlog',
    'matplotlib',
    'mmh3>=2.5.1',
    'networkx',
    'numpy',
    'pandas',
    'seaborn',
    'scipy',
    'xopen',
]

test_requirements = ["pytest", "mock"]

classifiers = ['Programming Language :: Python',
               'Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: 3.8',
               'Programming Language :: Python :: 3.9',
               'Programming Language :: Python :: 3.10',
               'Programming Language :: Python :: 3.11',
               'License :: OSI Approved :: MIT License',
               'Operating System :: OS Independent',
               'Development Status :: 4 - Beta',
               'Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'Intended Audience :: End Users/Desktop',
               'Topic :: Scientific/Engineering :: Bio-Informatics',
               'Topic :: Scientific/Engineering :: Chemistry',
               'Topic :: Software Development :: Libraries :: Python Modules'
               ]

setup(
    name='luna',
    version=version,
    packages=find_packages(exclude=['docs', 'tests']),
    python_requires='>=3.7, <4',
    install_requires=requirements,
    extras_require={
        'docs': ['sphinx >= 1.4', 'sphinx-rtd-theme']
    },
    include_package_data=True,
    description='LUNA: drug discovery toolkit',
    long_description=get_readme(),
    long_description_content_type='text/x-rst',
    url='https://github.com/keiserlab/LUNA',
    author='Alexandre Fassio',
    author_email='afassio@keiserlab.org',
    license='MIT',
    classifiers=classifiers,
    keywords='luna eifp fifp hifp ifp interaction fingerprint protein-ligand molecule docking',
    project_urls={
        'Bug Reports': 'https://github.com/keiserlab/LUNA/issues',
        'Source': 'https://github.com/keiserlab/LUNA/',
    },
    download_url='https://github.com/keiserlab/LUNA/tarball/v' + version,
    tests_require=test_requirements,

    entry_points={
        'console_scripts': [
            'run_luna = luna.run:main'
        ]
    }
)


from collections import defaultdict
import logging


logger = logging.getLogger()


def count_interaction_types(interactions, must_have_target=True, compounds=None, key_map={}):
    """Count the number of each type of interaction in ``interactions``.

    Parameters
    ----------
    interactions : iterable
            An iterable object containing a sequence of interactions (``InteractionType``).
    must_have_target : bool
            If True, count only interactions involving the target ligand. The default value is True.
    compounds: iterable, optional
            Only count interactions involving the compounds in ``compounds``.
            The default value is None, which implies that all compounds will be considered.
    key_map: dictionary, optional
            A dictionary to control which interactions to count and how to aggregate them.
            The keys are the interaction types to be considered and the values are the final interaction type,
            which can be used to aggregate interactions. If a value is None, the interaction will be ignored.

            For example, to aggregate all covalent interactions, ``key_map`` could be defined as follows:

               .. code-block:: python

                    key_map = {"Single bond": "Covalent bond",
                               "Double bond": "Covalent bond",
                               "Triple bond": "Covalent bond",
                               "Aromatic bond": "Covalent bond"}

            Now, if Ionic interactions should be ignored, ``key_map`` could be defined as follows:

                .. code-block:: python

                    key_map = {"Ionic": None}

    Returns
    -------
    interaction_types_count : dict
        The count of each interaction type.
    """
    if compounds:
        compounds = set(compounds)

    interaction_types_count = defaultdict(int)
    seen_pairs = set()
    for i in interactions:
        is_valid = False
        if not must_have_target and compounds is None:
            is_valid = True
        else:
            if must_have_target:
                if i.src_grp.has_target() or i.trgt_grp.has_target():
                    is_valid = True

            if compounds:
                if len(i.src_grp.compounds & compounds) > 0 or len(i.trgt_grp.compounds & compounds):
                    is_valid = True

        if is_valid:
            pair_key1 = (i.type, i.src_grp, i.trgt_grp)
            pair_key2 = (i.type, i.trgt_grp, i.src_grp)

            if pair_key1 in seen_pairs or pair_key2 in seen_pairs:
                continue

            seen_pairs.add(pair_key1)
            key = i.type
            if key_map:
                if i.type in key_map:
                    if key_map[i.type] is None:
                        continue
                    key = key_map[i.type]
                else:
                    key = i.type

            interaction_types_count[key] += 1
    return interaction_types_count
